// Copyright 2020 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include "ios/chrome/browser/passwords/ios_chrome_password_check_manager.h"

#include <memory>
#include <string>
#include <vector>

#include "base/bind.h"
#include "base/memory/scoped_refptr.h"
#include "base/strings/strcat.h"
#include "base/strings/string_number_conversions.h"
#include "base/strings/string_piece.h"
#include "base/strings/string_util.h"
#include "base/strings/utf_string_conversions.h"
#include "base/test/bind_test_util.h"
#include "base/test/scoped_feature_list.h"
#include "components/password_manager/core/browser/bulk_leak_check_service.h"
#include "components/password_manager/core/browser/mock_bulk_leak_check_service.h"
#include "components/password_manager/core/browser/mock_password_store.h"
#include "components/password_manager/core/browser/password_manager_test_utils.h"
#include "components/password_manager/core/browser/test_password_store.h"
#include "components/password_manager/core/common/password_manager_features.h"
#include "components/password_manager/core/common/password_manager_pref_names.h"
#include "components/prefs/pref_registry_simple.h"
#include "components/prefs/testing_pref_service.h"
#include "ios/chrome/browser/browser_state/chrome_browser_state.h"
#include "ios/chrome/browser/browser_state/test_chrome_browser_state.h"
#include "ios/chrome/browser/passwords/ios_chrome_bulk_leak_check_service_factory.h"
#include "ios/chrome/browser/passwords/ios_chrome_password_check_manager_factory.h"
#include "ios/chrome/browser/passwords/ios_chrome_password_store_factory.h"
#include "ios/web/public/test/web_task_environment.h"
#include "testing/gmock/include/gmock/gmock.h"
#include "testing/gtest/include/gtest/gtest.h"
#include "testing/platform_test.h"

#if !defined(__has_feature) || !__has_feature(objc_arc)
#error "This file requires ARC support."
#endif

namespace {
constexpr char kExampleCom[] = "https://example.com";

constexpr char kUsername1[] = "alice";

constexpr char kPassword1[] = "s3cre3t";

using autofill::PasswordForm;
using password_manager::BulkLeakCheckServiceInterface;
using password_manager::CredentialWithPassword;
using password_manager::MockBulkLeakCheckService;
using password_manager::CompromisedCredentials;
using password_manager::CompromiseType;
using password_manager::CompromiseTypeFlags;
using password_manager::IsLeaked;
using password_manager::LeakCheckCredential;
using password_manager::TestPasswordStore;
using ::testing::AllOf;
using ::testing::ElementsAre;
using ::testing::Field;
using ::testing::IsEmpty;
using ::testing::StrictMock;

using CompromisedCredentialsView =
    password_manager::CompromisedCredentialsManager::CredentialsView;

struct MockPasswordCheckManagerObserver
    : IOSChromePasswordCheckManager::Observer {
  MOCK_METHOD(void,
              CompromisedCredentialsChanged,
              (CompromisedCredentialsView),
              (override));
  MOCK_METHOD(void,
              PasswordCheckStatusChanged,
              (PasswordCheckState),
              (override));
};

scoped_refptr<TestPasswordStore> CreateAndUseTestPasswordStore(
    ChromeBrowserState* _browserState) {
  return base::WrapRefCounted(static_cast<password_manager::TestPasswordStore*>(
      IOSChromePasswordStoreFactory::GetInstance()
          ->SetTestingFactoryAndUse(
              _browserState,
              base::BindRepeating(&password_manager::BuildPasswordStore<
                                  web::BrowserState, TestPasswordStore>))
          .get()));
}

MockBulkLeakCheckService* CreateAndUseBulkLeakCheckService(
    ChromeBrowserState* _browserState) {
  return static_cast<MockBulkLeakCheckService*>(
      IOSChromeBulkLeakCheckServiceFactory::GetInstance()
          ->SetTestingFactoryAndUse(
              _browserState, base::BindLambdaForTesting([](web::BrowserState*) {
                return std::unique_ptr<KeyedService>(
                    std::make_unique<MockBulkLeakCheckService>());
              })));
}

CompromisedCredentials MakeCompromised(
    base::StringPiece signon_realm,
    base::StringPiece username,
    base::TimeDelta time_since_creation = base::TimeDelta(),
    CompromiseType compromise_type = CompromiseType::kLeaked) {
  return {
      std::string(signon_realm),
      base::ASCIIToUTF16(username),
      base::Time::Now() - time_since_creation,
      compromise_type,
  };
}

PasswordForm MakeSavedPassword(
    base::StringPiece signon_realm,
    base::StringPiece username,
    base::StringPiece password = kPassword1,
    base::StringPiece username_element = base::StringPiece()) {
  PasswordForm form;
  form.signon_realm = std::string(signon_realm);
  form.username_value = base::ASCIIToUTF16(username);
  form.password_value = base::ASCIIToUTF16(password);
  form.username_element = base::ASCIIToUTF16(username_element);
  return form;
}

// Creates matcher for a given compromised credential
auto ExpectCompromisedCredential(const std::string& signon_realm,
                                 const base::StringPiece& username,
                                 const base::StringPiece& password,
                                 base::TimeDelta elapsed_time_since_compromise,
                                 CompromiseTypeFlags compromise_type) {
  return AllOf(
      Field(&CredentialWithPassword::signon_realm, signon_realm),
      Field(&CredentialWithPassword::username, base::ASCIIToUTF16(username)),
      Field(&CredentialWithPassword::password, base::ASCIIToUTF16(password)),
      Field(&CredentialWithPassword::create_time,
            (base::Time::Now() - elapsed_time_since_compromise)),
      Field(&CredentialWithPassword::compromise_type, compromise_type));
}

class IOSChromePasswordCheckManagerTest : public PlatformTest {
 public:
  IOSChromePasswordCheckManagerTest()
      : browser_state_(TestChromeBrowserState::Builder().Build()),
        bulk_leak_check_service_(
            CreateAndUseBulkLeakCheckService(browser_state_.get())),
        store_(CreateAndUseTestPasswordStore(browser_state_.get())) {
    manager_ = IOSChromePasswordCheckManagerFactory::GetForBrowserState(
        browser_state_.get());
    scoped_feature_list_.InitAndEnableFeature(
        password_manager::features::kPasswordCheck);
  }

  void RunUntilIdle() { task_env_.RunUntilIdle(); }

  ChromeBrowserState* browser_state() { return browser_state_.get(); }
  TestPasswordStore& store() { return *store_; }
  MockBulkLeakCheckService* service() { return bulk_leak_check_service_; }
  IOSChromePasswordCheckManager& manager() { return *manager_; }

 private:
  base::test::ScopedFeatureList scoped_feature_list_;
  web::WebTaskEnvironment task_env_{
      web::WebTaskEnvironment::Options::DEFAULT,
      base::test::TaskEnvironment::TimeSource::MOCK_TIME};
  std::unique_ptr<ChromeBrowserState> browser_state_;
  MockBulkLeakCheckService* bulk_leak_check_service_;
  scoped_refptr<TestPasswordStore> store_;
  scoped_refptr<IOSChromePasswordCheckManager> manager_;
};
}  // namespace

// Sets up the password store with a password and compromised
// credential. Verifies that the result is matching expectation.
TEST_F(IOSChromePasswordCheckManagerTest, GetCompromisedCredentials) {
  store().AddLogin(MakeSavedPassword(kExampleCom, kUsername1));

  store().AddCompromisedCredentials(
      MakeCompromised(kExampleCom, kUsername1, base::TimeDelta::FromMinutes(1),
                      CompromiseType::kLeaked));
  RunUntilIdle();
  EXPECT_THAT(
      manager().GetCompromisedCredentials(),
      ElementsAre(ExpectCompromisedCredential(
          kExampleCom, kUsername1, kPassword1, base::TimeDelta::FromMinutes(1),
          CompromiseTypeFlags::kCredentialLeaked)));
}

// Test that we don't create an entry in the password store if IsLeaked is
// false.
TEST_F(IOSChromePasswordCheckManagerTest, NoLeakedFound) {
  store().AddLogin(MakeSavedPassword(kExampleCom, kUsername1, kPassword1));
  RunUntilIdle();

  static_cast<BulkLeakCheckServiceInterface::Observer*>(&manager())
      ->OnCredentialDone(LeakCheckCredential(base::ASCIIToUTF16(kUsername1),
                                             base::ASCIIToUTF16(kPassword1)),
                         IsLeaked(false));
  RunUntilIdle();

  EXPECT_THAT(manager().GetCompromisedCredentials(), IsEmpty());
}

// Test that a found leak creates a compromised credential in the password
// store.
TEST_F(IOSChromePasswordCheckManagerTest, OnLeakFoundCreatesCredential) {
  store().AddLogin(MakeSavedPassword(kExampleCom, kUsername1, kPassword1));
  RunUntilIdle();

  static_cast<BulkLeakCheckServiceInterface::Observer*>(&manager())
      ->OnCredentialDone(LeakCheckCredential(base::ASCIIToUTF16(kUsername1),
                                             base::ASCIIToUTF16(kPassword1)),
                         IsLeaked(true));
  RunUntilIdle();

  EXPECT_THAT(
      manager().GetCompromisedCredentials(),
      ElementsAre(ExpectCompromisedCredential(
          kExampleCom, kUsername1, kPassword1, base::TimeDelta::FromMinutes(0),
          CompromiseTypeFlags::kCredentialLeaked)));
}

// Verifies that the case where the user has no saved passwords is reported
// correctly.
TEST_F(IOSChromePasswordCheckManagerTest, GetPasswordCheckStatusNoPasswords) {
  EXPECT_EQ(PasswordCheckState::kNoPasswords,
            manager().GetPasswordCheckState());
}

// Verifies that the case where the user has saved passwords is reported
// correctly.
TEST_F(IOSChromePasswordCheckManagerTest, GetPasswordCheckStatusIdle) {
  store().AddLogin(MakeSavedPassword(kExampleCom, kUsername1));
  RunUntilIdle();

  EXPECT_EQ(PasswordCheckState::kIdle, manager().GetPasswordCheckState());
}

// Checks that the default kLastTimePasswordCheckCompleted pref value is
// treated as no completed run yet.
TEST_F(IOSChromePasswordCheckManagerTest,
       LastTimePasswordCheckCompletedNotSet) {
  EXPECT_EQ(base::Time(), manager().GetLastPasswordCheckTime());
}

// Checks that a non-default kLastTimePasswordCheckCompleted pref value is
// treated as a completed run.
TEST_F(IOSChromePasswordCheckManagerTest, LastTimePasswordCheckCompletedIsSet) {
  base::Time expected = base::Time::Now() - base::TimeDelta::FromMinutes(5);
  browser_state()->GetPrefs()->SetDouble(
      password_manager::prefs::kLastTimePasswordCheckCompleted,
      expected.ToDoubleT());

  EXPECT_THAT(expected, manager().GetLastPasswordCheckTime());
}

// Checks that a transition into the idle state after starting a check results
// in resetting the kLastTimePasswordCheckCompleted pref to the current time.
TEST_F(IOSChromePasswordCheckManagerTest, LastTimePasswordCheckCompletedReset) {
  manager().StartPasswordCheck();
  RunUntilIdle();

  static_cast<BulkLeakCheckServiceInterface::Observer*>(&manager())
      ->OnStateChanged(BulkLeakCheckServiceInterface::State::kIdle);

  EXPECT_THAT(base::Time::Now(), manager().GetLastPasswordCheckTime());
}

// Tests whether adding and removing an observer works as expected.
TEST_F(IOSChromePasswordCheckManagerTest,
       NotifyObserversAboutCompromisedCredentialChanges) {
  store().AddLogin(MakeSavedPassword(kExampleCom, kUsername1));

  RunUntilIdle();
  StrictMock<MockPasswordCheckManagerObserver> observer;
  manager().AddObserver(&observer);

  // Adding a compromised credential should notify observers.
  EXPECT_CALL(
      observer,
      CompromisedCredentialsChanged(ElementsAre(ExpectCompromisedCredential(
          kExampleCom, kUsername1, kPassword1, base::TimeDelta::FromMinutes(1),
          CompromiseTypeFlags::kCredentialLeaked))));
  store().AddCompromisedCredentials(MakeCompromised(
      kExampleCom, kUsername1, base::TimeDelta::FromMinutes(1)));
  RunUntilIdle();

  // After an observer is removed it should no longer receive notifications.
  manager().RemoveObserver(&observer);
  EXPECT_CALL(observer, CompromisedCredentialsChanged).Times(0);
  store().AddCompromisedCredentials(
      MakeCompromised(kExampleCom, kUsername1, base::TimeDelta::FromMinutes(1),
                      CompromiseType::kPhished));
  RunUntilIdle();
}

// Tests whether adding and removing an observer works as expected.
TEST_F(IOSChromePasswordCheckManagerTest, NotifyObserversAboutStateChanges) {
  store().AddLogin(MakeSavedPassword(kExampleCom, kUsername1));
  RunUntilIdle();
  StrictMock<MockPasswordCheckManagerObserver> observer;
  manager().AddObserver(&observer);

  EXPECT_CALL(observer, PasswordCheckStatusChanged(PasswordCheckState::kIdle));
  static_cast<BulkLeakCheckServiceInterface::Observer*>(&manager())
      ->OnStateChanged(BulkLeakCheckServiceInterface::State::kIdle);

  RunUntilIdle();

  EXPECT_EQ(PasswordCheckState::kIdle, manager().GetPasswordCheckState());

  // After an observer is removed it should no longer receive notifications.
  manager().RemoveObserver(&observer);
  EXPECT_CALL(observer, PasswordCheckStatusChanged).Times(0);
  static_cast<BulkLeakCheckServiceInterface::Observer*>(&manager())
      ->OnStateChanged(BulkLeakCheckServiceInterface::State::kRunning);
  RunUntilIdle();
}
